// GITVERSION: texstuff-1.4.0-102-g1de664bd
#ifndef EVAL_HPP
#define EVAL_HPP
#include "aedtools/evalbase.hpp"
#include <forward_list>
#include <sstream>
#include <algorithm>
using namespace std;

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class inthasher_t {
public:
  virtual void reset()=0;
  virtual void hash(int x)=0;
  virtual int val()=0;
  virtual ~inthasher_t() {}
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class aedhasher_t : public inthasher_t {
public:
  aed::hasher_t aedh;
  aedhasher_t() { reset(); }
  void reset() { aedh.reset(); }
  void hash(int x) { aedh.hash(x); }
  int val() { return aedh.val(); }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class dumbhasher_t : public inthasher_t {
public:
  aed::hasher_t aedh;
  dumbhasher_t() { }
  void reset() { }
  void hash(int x) { }
  int val() { return 0; }
};

void mksumhasher(unique_ptr<inthasher_t> &p);
void mkcubhasher(unique_ptr<inthasher_t> &p);

int count_different(vector<vector<int>> &vectors,
                    inthasher_t &hasher);
typedef vector<set<int>> graph_t;
void isograph(graph_t &g1,graph_t &g2,
              inthasher_t &hasher,vector<int> &perm);

void mkgraph(json &data,graph_t &g) {
  g.clear();
  for (auto &z : data) {
    set<int> s;
    for (auto &x : z)
      s.insert(x.get<int>());
    g.push_back(s);
  }
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void print_graph2(graph_t &g,const char *label=NULL) {
  if (label) printf("%s -------\n",label);
  int nv=g.size();
  for (int j=0; j<nv; j++) {
    printf("%d: ",j);
    set<int> &ngbr = g[j];
    for (auto &q: ngbr) printf("%d ",q);
    printf("\n");
  }
}

namespace aed {

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
  //   FUNCTIONS FOR THIS EXAM
  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
  // END FUNCTIONS FOR THIS EXAM
  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> ISINVERSE
#undef EJERC
#undef EVAL
#define EJERC 1
#define EVAL eval1_t
  
  class EVAL : public eval_base_t {
  public:
    EVAL() { dumptests=0; ejerc=EJERC; testfile="./sumhasher.json"; }
    void run_case(json &data,json &outuser) {
      unique_ptr<inthasher_t> p;
      mksumhasher(p);
      vector<int> w = data["v"];
      p->reset();
      for (auto &x : w) p->hash(x);
      outuser["hash"] = p->val();
    }

    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }

    void generate_case(randomg_t &rnd,json &datain) {
      int m = 1+rnd.rand()%10;
      json z;
      for (int j=0; j<m; j++) z.push_back(rnd.rand()%1000);
      datain["v"] = z;
    }
  };

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> ISINVERSE
#undef EJERC
#undef EVAL
#define EJERC 2
#define EVAL eval2_t
  class EVAL : public eval_base_t {
  public:
    EVAL() { dumptests=0; ejerc=EJERC; testfile="./cubhasher.json"; }
    void run_case(json &data,json &outuser) {
      unique_ptr<inthasher_t> p;
      mkcubhasher(p);
      vector<int> w = data["v"];
      p->reset();
      for (auto &x : w) p->hash(x);
      outuser["hash"] = p->val();
    }

    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }

    void generate_case(randomg_t &rnd,json &datain) {
      int m = 1+rnd.rand()%10;
      json z;
      for (int j=0; j<m; j++) z.push_back(rnd.rand()%1000);
      datain["v"] = z;
    }
  };

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> ISINVERSE
#undef EJERC
#undef EVAL
#define EJERC 3
#define EVAL eval3_t
  class EVAL : public eval_base_t {
  public:
    EVAL() { dumptests=0; ejerc=EJERC; testfile="./countdiff.json"; }
    void run_case(json &data,json &outuser) {
      unique_ptr<inthasher_t> p;
      mkcubhasher(p);
      vector<vector<int>> vectors;
      for (auto &x : data["vectors"]) {
        vector<int> v = x;
        vectors.push_back(v);
      }
      outuser["count"] = count_different(vectors,*p.get());
    }

    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }

    void generate_case(randomg_t &rnd,json &datain) {
      int nvec=100;
      json vecs;
      for (int j=0; j<nvec; j++) {
        int m=2+rnd.rand()%2;
        json z;
        for (int j=0; j<m; j++) {
          int x = rnd.rand()%4;
          z.push_back(x);
        }
        vecs.push_back(z);
      }
      datain["vectors"] = vecs;
    }
  };

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> ISOGRAPH
#undef EJERC
#undef EVAL
#define EJERC 4
#define EVAL eval4_t
  class EVAL : public eval_base_t {
  public:
    EVAL() { dumptests=0; ejerc=EJERC; testfile="./isograph.json"; }
    void run_case(json &data,json &outuser) {
      graph_t g1,g2;
      mkgraph(data["g1"],g1);
      mkgraph(data["g2"],g2);
      aedhasher_t hasher;
      vector<int> perm;
      isograph(g1,g2,hasher,perm);
      outuser["perm"] = perm;
    }

    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }

    void generate_case(randomg_t &rnd,json &datain) {
      assert(0); // NOT SUPPOSED TO BE CALLED!! 
    }
  };
  
}

#undef CSTR

#endif
