// GITVERSION: texstuff-1.4.0-224-gbf3b2aba
#ifndef EVAL_HPP
#define EVAL_HPP
#include "aedtools/evalbase.hpp"
#include <forward_list>
#include <sstream>
#include <algorithm>
#include <Eigen/Dense>

using namespace std;
using namespace Eigen;

#define ESHV(x) cout << #x << endl << x << endl

int maxleaf(json &tree);
int maxdepth(json &tree);
void xcloser(MatrixXd &vert,VectorXd &x,VectorXd &y);

namespace aed {

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
  //   FUNCTIONS FOR THIS EXAM
  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>


  void json2eigen(json &j,MatrixXd &A,
                  int m,int n,int rowwise=1) {
    // ESHV(j);
    int sz =m*n;
    A.resize(sz,1);
    assert(int(j.size())==sz);
    int l=0;
    for (auto &x : j) A(l++,0) = x.get<double>();
    if (rowwise) {
      A.resize(n,m);
      A.transposeInPlace();
    } else {
      A.resize(m,n);
    }
  }

  void json2eigen(json &j,VectorXd &x,int m) {
    MatrixXd A(m,1);
    json2eigen(j,A,m,1);
    x=A.col(0);
  }
  
  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
  // END FUNCTIONS FOR THIS EXAM
  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> MAXLEAF
#undef EJERC
#undef EVAL
#define EJERC 1
#define EVAL eval1_t
  
  class EVAL : public eval_base_t {
  public:
    EVAL() { dumptests=0; ejerc=EJERC; testfile="./maxleaf.json"; }
    void run_case(json &data,json &outuser) {
      outuser["retval"] = maxleaf(data["T"]);
    }

    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }

    void generate_case(randomg_t &rnd,json &datain) {
      tree<int> T;
      int m = 5+rnd.rand()%10;
      int n = 5+rnd.rand()%10;
      make_random_tree2(T,100,m,n,rnd);
      datain["T"] = tree2json(T);
    }
  };

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> MAXDEPTH
#undef EJERC
#undef EVAL
#define EJERC 2
#define EVAL eval2_t
  
  class EVAL : public eval_base_t {
  public:
    EVAL() { dumptests=0; ejerc=EJERC; testfile="./maxdepth.json"; }
    void run_case(json &data,json &outuser) {
      outuser["retval"] = maxdepth(data["T"]);
    }

    int check_case(json &datain,
                   json &outref,json &outuser) {
      return outref==outuser;
    }

    void generate_case(randomg_t &rnd,json &datain) {
      tree<int> T;
      int m = 5+rnd.rand()%10;
      int n = 5+rnd.rand()%10;
      make_random_tree2(T,100,m,n,rnd);
      datain["T"] = tree2json(T);
    }
  };

  //---:---<*>---:---<*>---:---<*>---:---<*>---:---<*> XCLOSER
#undef EJERC
#undef EVAL
#define EJERC 3
#define EVAL eval3_t
  
  class EVAL : public eval_base_t {
  public:
    EVAL() { dumptests=0; ejerc=EJERC; testfile="./xcloser.json"; }
    void run_case(json &data,json &outuser) {
      int m=data["verts"].size();
      int ndim = (m==6? 2 : m==12? 3 : -1);
      MatrixXd verts;
      json2eigen(data["verts"],verts,ndim,ndim+1,0);
      // ESHV(verts);
      VectorXd x(ndim),y(ndim);
      json2eigen(data["x"],x,ndim);
      // ESHV(x);
      xcloser(verts,x,y);
      json jy;
      for (int j=0; j<ndim; j++)
        jy.push_back(y(j));
      outuser["y"].swap(jy);
    }

    int check_case(json &datain,
                   json &outref,json &outuser) {
      VectorXd xref,xuser;
      int ndim=outref["y"].size();
      json2eigen(outref["y"],xref,ndim);
      json2eigen(outuser["y"],xuser,ndim);
      return (xref-xuser).norm()<1e-8;
    }

    void generate_case(randomg_t &rnd,json &datain) {
      datain.clear();
      datain["verts"] = {0,0,1,0,0,1};
      int N=100;
      for (int j=0; j<2; j++) {
        double z=-0.2+(2.0/N)*(rand()%N);
        datain["x"].push_back(z);
      }
    }
  };

}

#undef CSTR

#endif
